###############################################################################
# ProteusLib Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/nawi-hub/proteuslib/"
#
###############################################################################
"""
Artificial seawater properties comprising Na+, Cl-, Ca_2+, SO4_2-, Mg_2+
Equation of state: eNRTL

eNRTL property configuration dicts for synthetic hard water based on [1]

References:

[1] Islam, R.I., et al., Molecular thermodynamics for scaling prediction: Case
of membrane distillation, Separation and Purification Technology, 2021,
Vol. 276.
"""

from pyomo.environ import Param, units as pyunits

from idaes.core import (LiquidPhase,
                        AqueousPhase,
                        Solvent,
                        Apparent,
                        Anion,
                        Cation,
                        MaterialFlowBasis)
from idaes.generic_models.properties.core.eos.enrtl import ENRTL
from idaes.generic_models.properties.core.eos.enrtl_reference_states import \
    Symmetric, Unsymmetric
from idaes.generic_models.properties.core.generic.generic_property import (
        StateIndex)
from idaes.generic_models.properties.core.state_definitions import FpcTP
from idaes.generic_models.properties.core.pure.electrolyte import \
    relative_permittivity_constant
import idaes.logger as idaeslog
# Set up logger
_log = idaeslog.getLogger(__name__)

class VolMolH2O():
    def build_parameters(b):
        b.vol_mol_pure = Param(initialize=18e-6,
                               units=pyunits.m**3/pyunits.mol,
                               mutable=True)

    def return_expression(b, cobj, T):
        return cobj.vol_mol_pure

class VolMolNaCl():
    def build_parameters(b):
        b.vol_mol_pure = Param(initialize=58.44e-6,
                               units=pyunits.m**3/pyunits.mol,
                               mutable=True)

    def return_expression(b, cobj, T):
        return cobj.vol_mol_pure

class VolMolNa2SO4():
    def build_parameters(b):
        b.vol_mol_pure = Param(initialize=142.04e-6,
                               units=pyunits.m**3/pyunits.mol,
                               mutable=True)

    def return_expression(b, cobj, T):
        return cobj.vol_mol_pure


class VolMolCaCl2():
    def build_parameters(b):
        b.vol_mol_pure = Param(initialize=110.98e-6,
                               units=pyunits.m**3/pyunits.mol,
                               mutable=True)

    def return_expression(b, cobj, T):
        return cobj.vol_mol_pure

class VolMolCaSO4():
    def build_parameters(b):
        b.vol_mol_pure = Param(initialize=136.14e-6,
                               units=pyunits.m**3/pyunits.mol,
                               mutable=True)

    def return_expression(b, cobj, T):
        return cobj.vol_mol_pure

class VolMolMgCl2():
    def build_parameters(b):
        b.vol_mol_pure = Param(initialize=95.21e-6,
                               units=pyunits.m**3/pyunits.mol,
                               mutable=True)

    def return_expression(b, cobj, T):
        return cobj.vol_mol_pure

class VolMolMgSO4():
    def build_parameters(b):
        b.vol_mol_pure = Param(initialize=120.37e-6,
                               units=pyunits.m**3/pyunits.mol,
                               mutable=True)

    def return_expression(b, cobj, T):
        return cobj.vol_mol_pure

configuration = {
    "components": {
        "H2O": {"type": Solvent,
                "vol_mol_liq_comp": VolMolH2O,
                "relative_permittivity_liq_comp":
                    relative_permittivity_constant,
                "parameter_data": {
                    "mw": (18E-3, pyunits.kg/pyunits.mol),
                    "relative_permittivity_liq_comp": 78.54}},
        "NaCl": {"type": Apparent,
                 "dissociation_species": {"Na_+": 1, "Cl_-": 1},
                 "vol_mol_liq_comp": VolMolNaCl,
                 "parameter_data": {
                     "mw": (58.44E-3, pyunits.kg / pyunits.mol)
                 }},
        "Na2SO4": {"type": Apparent,
                   "dissociation_species": {"Na_+": 2, "SO4_2-": 1},
                   "vol_mol_liq_comp": VolMolNa2SO4,
                   "parameter_data": {
                       "mw": (142.04E-3, pyunits.kg / pyunits.mol)
                   }},
        "CaCl2": {"type": Apparent,
                  "dissociation_species": {"Ca_2+": 1, "Cl_-": 2},
                  "vol_mol_liq_comp": VolMolCaCl2,
                  "parameter_data": {
                      "mw": (110.98E-3, pyunits.kg / pyunits.mol)
                  }},
        "CaSO4": {"type": Apparent,
                  "dissociation_species": {"Ca_2+": 1, "SO4_2-": 1},
                  "vol_mol_liq_comp": VolMolCaSO4,
                  "parameter_data": {
                      "mw": (136.14E-3, pyunits.kg / pyunits.mol)
                  }},
        "MgCl2": {"type": Apparent,
                  "dissociation_species": {"Mg_2+": 1, "Cl_-": 2},
                  "vol_mol_liq_comp": VolMolMgCl2,
                  "parameter_data": {
                      "mw": (95.21E-3, pyunits.kg / pyunits.mol)
                  }},
        "MgSO4": {"type": Apparent,
                  "dissociation_species": {"Mg_2+": 1, "SO4_2-": 1},
                  "vol_mol_liq_comp": VolMolMgSO4,
                  "parameter_data": {
                      "mw": (120.37E-3, pyunits.kg / pyunits.mol)
                  }},
        "Na_+": {"type": Cation,
                 "charge": +1,
                 "parameter_data":{
                     "mw": 23e-3
                 }},
        "Ca_2+": {"type": Cation,
                  "charge": +2,
                  "parameter_data": {
                      "mw": 40e-3
                  }},
        "Mg_2+": {"type": Cation,
                  "charge": +2,
                  "parameter_data":{
                      "mw": 24.3e-3
                  }},
        "Cl_-": {"type": Anion,
                 "charge": -1,
                 "parameter_data": {
                     "mw": 35.45e-3
                 }},
        "SO4_2-": {"type": Anion,
                   "charge": -2,
                   "parameter_data": {
                       "mw": 96.06e-3
                   }}},
    "phases": {
        "Liq": {"type": AqueousPhase,
                "equation_of_state": ENRTL,
                "equation_of_state_options": {
                    "reference_state": Symmetric}}},
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},
    "state_definition": FpcTP,
    "state_components": StateIndex.true,
    "reaction_basis": MaterialFlowBasis.molar,
    "pressure_ref": 1e5,
    "temperature_ref": 298.15,
    "parameter_data": {
        "Liq_tau": {  # Table 1 [1]
            ('H2O', 'Na_+, Cl_-'): 8.866,
            ('Na_+, Cl_-', 'H2O'): -4.451,
            ('H2O', 'Ca_2+, Cl_-'): 10.478,
            ('Ca_2+, Cl_-', 'H2O'): -5.231,
            ('H2O', 'Mg_2+, Cl_-'): 10.854,
            ('Mg_2+, Cl_-', 'H2O'): -5.409,
            ('H2O', 'Na_+, SO4_2-'): 8.012,
            ('Na_+, SO4_2-', 'H2O'): -3.903,
            ('H2O', 'Ca_2+, SO4_2-'): 6.932,
            ('Ca_2+, SO4_2-', 'H2O'): -3.466,
            ('H2O', 'Mg_2+, SO4_2-'): 8.808,
            ('Mg_2+, SO4_2-', 'H2O'): -4.383,
            ('Na_+, Cl_-', 'Ca_2+, Cl_-'): -0.468,
            ('Ca_2+, Cl_-', 'Na_+, Cl_-'): 0.41,
            ('Na_+, Cl_-', 'Mg_2+, Cl_-'): -0.328,
            ('Mg_2+, Cl_-', 'Na_+, Cl_-'): -0.981,
            ('Ca_2+, Cl_-', 'Mg_2+, Cl_-'): 0.22,
            ('Mg_2+, Cl_-', 'Ca_2+, Cl_-'): 0.322,
            ('Na_+, SO4_2-', 'Ca_2+, SO4_2-'): -0.761,
            ('Ca_2+, SO4_2-', 'Na_+, SO4_2-'): 0.368,
            ('Na_+, SO4_2-', 'Mg_2+, SO4_2-'): -0.327,
            ('Mg_2+, SO4_2-', 'Na_+, SO4_2-'): 0.799,
            ('Ca_2+, SO4_2-', 'Mg_2+, SO4_2-'): 0,
            ('Mg_2+, SO4_2-', 'Ca_2+, SO4_2-'): 0.383,
            ('Na_+, Cl_-', 'Na_+, SO4_2-'): 0.803,
            ('Na_+, SO4_2-', 'Na_+, Cl_-'): -0.634,
            ('Ca_2+, Cl_-', 'Ca_2+, SO4_2-'): 0,
            ('Ca_2+, SO4_2-', 'Ca_2+, Cl_-'): -0.264,
            ('Mg_2+, Cl_-', 'Mg_2+, SO4_2-'): -0.707,
            ('Mg_2+, SO4_2-', 'Mg_2+, Cl_-'): -0.841
            }},
    "default_scaling_factors": {("flow_mol_phase_comp", ("Liq", "Na_+")): 1e1,
                                ("flow_mol_phase_comp", ("Liq", "Ca_2+")): 1e3,
                                ("flow_mol_phase_comp", ("Liq", "Mg_2+")): 1e2,
                                ("flow_mol_phase_comp", ("Liq", "SO4_2-")): 1e2,
                                ("flow_mol_phase_comp", ("Liq", "Cl_-")): 1e1,
                                ("flow_mol_phase_comp", ("Liq", "H2O")): 1e-1,
                                ("mole_frac_comp", "Na_+"): 1e2,
                                ("mole_frac_comp", "Ca_2+"): 1e4,
                                ("mole_frac_comp", "Mg_2+"): 1e3,
                                ("mole_frac_comp", "SO4_2-"): 1e3,
                                ("mole_frac_comp", "Cl_-"): 1e2,
                                ("mole_frac_comp", "H2O"): 1,
                                ("mole_frac_phase_comp", ("Liq", "Na_+")): 1e2,
                                ("mole_frac_phase_comp", ("Liq", "Ca_2+")): 1e4,
                                ("mole_frac_phase_comp", ("Liq", "Mg_2+")): 1e3,
                                ("mole_frac_phase_comp", ("Liq", "SO4_2-")): 1e3,
                                ("mole_frac_phase_comp", ("Liq", "Cl_-")): 1e2,
                                ("mole_frac_phase_comp", ("Liq", "H2O")): 1,
                                ("flow_mol_phase_comp_apparent", ("Liq", "NaCl")): 1e1,
                                ("flow_mol_phase_comp_apparent", ("Liq", "Na2SO4")): 1e2,
                                ("flow_mol_phase_comp_apparent", ("Liq", "CaCl2")): 1e2,
                                ("flow_mol_phase_comp_apparent", ("Liq", "CaSO4")): 1e3,
                                ("flow_mol_phase_comp_apparent", ("Liq", "MgCl2")): 1e2,
                                ("flow_mol_phase_comp_apparent", ("Liq", "MgSO4")): 1e3,
                                ("flow_mol_phase_comp_apparent", ("Liq", "H2O")): 1e-1,
                                ("mole_frac_phase_comp_apparent", ("Liq", "NaCl")): 1e3,  # TODO: these seem to be 1 orders of magnitude too low
                                ("mole_frac_phase_comp_apparent", ("Liq", "Na2SO4")): 1e4,
                                ("mole_frac_phase_comp_apparent", ("Liq", "CaCl2")): 1e4,
                                ("mole_frac_phase_comp_apparent", ("Liq", "CaSO4")): 1e5,
                                ("mole_frac_phase_comp_apparent", ("Liq", "MgCl2")): 1e4,
                                ("mole_frac_phase_comp_apparent", ("Liq", "MgSO4")): 1e5,
                                ("mole_frac_phase_comp_apparent", ("Liq", "H2O")): 1,
                                }
}



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""Simple zero order separator examples"""

from pyomo.environ import ConcreteModel, Constraint
from idaes.core import FlowsheetBlock
# from idaes.generic_models.unit_models import Separator  # replaced separator
from idaes.generic_models.unit_models.separator import SplittingType, EnergySplittingType
from idaes.core.util.scaling import calculate_scaling_factors, set_scaling_factor, constraint_scaling_transform
from watertap.examples.flowsheets.full_treatment_train.model_components import property_models
from watertap.examples.flowsheets.full_treatment_train.util import solve_with_user_scaling, check_dof
from watertap.examples.flowsheets.full_treatment_train.model_components import Separator


def build_SepRO(m, base='TDS'):
    """
    Builds RO model based on the IDAES separator.
    Requires prop_TDS property package.
    """
    prop = property_models.get_prop(m, base=base)

    m.fs.RO = Separator(default={
        "property_package": prop,
        "outlet_list": ['retentate', 'permeate'],
        "split_basis": SplittingType.componentFlow,
        "energy_split_basis": EnergySplittingType.equal_temperature})

    # specify
    if base == 'TDS':
        m.fs.RO.split_fraction[0, 'permeate', 'H2O'].fix(0.5)
        m.fs.RO.split_fraction[0, 'permeate', 'TDS'].fix(0.01)
    else:
        raise ValueError('Unexpected property base {base} provided to build_SepRO'
                         ''.format(base=base))

    # scale
    set_scaling_factor(m.fs.RO.split_fraction, 1)  # TODO: IDAES should set these scaling factors by default
    constraint_scaling_transform(m.fs.RO.sum_split_frac[0.0, 'H2O'], 1)
    constraint_scaling_transform(m.fs.RO.sum_split_frac[0.0, 'TDS'], 1)


def build_SepNF(m, base='ion'):
    """
    Builds NF model based on the IDAES separator for a specified property base.
    Requires prop_ion or prop_salt property package.
    """
    prop = property_models.get_prop(m, base=base)

    m.fs.NF = Separator(default={
        "property_package": prop,
        "outlet_list": ['retentate', 'permeate'],
        "split_basis": SplittingType.componentFlow,
        "energy_split_basis": EnergySplittingType.equal_temperature})

    # specify
    if base == 'ion':
        m.fs.NF.split_fraction[0, 'permeate', 'H2O'].fix(0.9)
        m.fs.NF.split_fraction[0, 'permeate', 'Na'].fix(0.9)
        m.fs.NF.split_fraction[0, 'permeate', 'Ca'].fix(0.1)
        m.fs.NF.split_fraction[0, 'permeate', 'Mg'].fix(0.1)
        m.fs.NF.split_fraction[0, 'permeate', 'SO4'].fix(0.1)
        # Cl split fraction determined through electro-neutrality for the retentate
        charge_dict = {'Na': 1, 'Ca': 2, 'Mg': 2, 'SO4': -2, 'Cl': -1}
        m.fs.NF.EN_out = Constraint(
            expr=0 ==
                 sum(charge_dict[j] * m.fs.NF.retentate_state[0].flow_mol_phase_comp['Liq', j]
                     for j in charge_dict))
        constraint_scaling_transform(m.fs.NF.EN_out, 1)
    elif base == 'salt':
        m.fs.NF.split_fraction[0, 'permeate', 'H2O'].fix(0.9)
        m.fs.NF.split_fraction[0, 'permeate', 'NaCl'].fix(0.9)
        m.fs.NF.split_fraction[0, 'permeate', 'CaSO4'].fix(0.1)
        m.fs.NF.split_fraction[0, 'permeate', 'MgSO4'].fix(0.1)
        m.fs.NF.split_fraction[0, 'permeate', 'MgCl2'].fix(0.2)

    # scale
    set_scaling_factor(m.fs.NF.split_fraction, 1)  # TODO: IDAES should set these scaling factors by default
    if base == 'ion':
        constraint_scaling_transform(m.fs.NF.sum_split_frac[0.0, 'H2O'], 1)
        constraint_scaling_transform(m.fs.NF.sum_split_frac[0.0, 'Na'], 1)
        constraint_scaling_transform(m.fs.NF.sum_split_frac[0.0, 'Ca'], 1)
        constraint_scaling_transform(m.fs.NF.sum_split_frac[0.0, 'Mg'], 1)
        constraint_scaling_transform(m.fs.NF.sum_split_frac[0.0, 'SO4'], 1)
        constraint_scaling_transform(m.fs.NF.sum_split_frac[0.0, 'Cl'], 1)
    elif base == 'salt':
        constraint_scaling_transform(m.fs.NF.sum_split_frac[0.0, 'H2O'], 1)
        constraint_scaling_transform(m.fs.NF.sum_split_frac[0.0, 'NaCl'], 1)
        constraint_scaling_transform(m.fs.NF.sum_split_frac[0.0, 'CaSO4'], 1)
        constraint_scaling_transform(m.fs.NF.sum_split_frac[0.0, 'MgSO4'], 1)
        constraint_scaling_transform(m.fs.NF.sum_split_frac[0.0, 'MgCl2'], 1)


def solve_SepRO(base='TDS'):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    property_models.build_prop(m, base=base)
    build_SepRO(m, base=base)
    property_models.specify_feed(m.fs.RO.mixed_state[0], base=base)

    check_dof(m)
    calculate_scaling_factors(m)
    solve_with_user_scaling(m)

    m.fs.RO.inlet.display()
    m.fs.RO.permeate.display()
    m.fs.RO.retentate.display()

    return m


def solve_SepNF(base='ion'):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    property_models.build_prop(m, base=base)
    build_SepNF(m, base=base)
    property_models.specify_feed(m.fs.NF.mixed_state[0], base=base)

    m.fs.NF.mixed_state[0].mass_frac_phase_comp  # touching for tests
    check_dof(m)
    calculate_scaling_factors(m)
    solve_with_user_scaling(m)

    m.fs.NF.inlet.display()
    m.fs.NF.permeate.display()
    m.fs.NF.retentate.display()

    return m


if __name__ == "__main__":
    solve_SepRO(base='TDS')
    solve_SepNF(base='ion')
    solve_SepNF(base='salt')



"""
Commands for Electrolyte Database
"""
# stdlib
import json
import logging
import pathlib
import sys
import tempfile

# third-party
import click
from json_schema_for_humans import generate as schema_gen
from json_schema_for_humans.generation_configuration import GenerationConfiguration
from pymongo import MongoClient
from pymongo.errors import ConnectionFailure

# package
from .db_api import ElectrolyteDB
from .validate import validate, ValidationError
from .schemas import schemas as edb_schemas

_log = logging.getLogger(__name__)


def get_edb_data(filename: str) -> pathlib.Path:
    """Get an installed electrolyte DB data file `filename`.

    Args:
        filename: File to get

    Returns:
        Path object for file.
    """
    from watertap import _ROOT

    return _ROOT / "edb" / "data" / filename


def level_from_verbosity(vb):
    level = 0  # for pycharm
    if vb >= 3:
        level = logging.DEBUG
    elif vb == 2:
        level = logging.INFO
    elif vb == 1:
        level = logging.WARN
    elif vb == 0:
        level = logging.ERROR
    elif vb == -1:
        level = logging.FATAL
    elif vb <= -2:
        level = logging.FATAL + 1
    return level


def _connect(url, db):
    """Connect to Mongo at given URL and database."""
    _log.info(f"Begin: Connect to MongoDB at: {url}/{db}")
    db = ElectrolyteDB(url=url, db=db)
    _log.info(f"End: Connect to MongoDB at: {url}/{db}")
    return db


@click.group()
@click.version_option(version=None)
@click.option(
    "--verbose",
    "-v",
    count=True,
    help="Increase verbosity. Show warnings if given once, "
    "then info, and then debugging messages.",
)
@click.option(
    "--quiet",
    "-q",
    count=True,
    help="Increase quietness. If given once, "
    "only show critical messages. If "
    "given twice, show no messages.",
)
def command_base(verbose, quiet):
    log_root = logging.getLogger("watertap")
    if quiet > 0 and verbose > 0:
        raise click.BadArgumentUsage("Options for verbosity and quietness conflict")
    if verbose > 0:
        _h = logging.StreamHandler()
        _h.setFormatter(
            logging.Formatter("%(asctime)s %(levelname)-7s %(name)s: %(message)s"))
        log_root.addHandler(_h)
        log_root.setLevel(level_from_verbosity(verbose))
    else:
        log_root.setLevel(level_from_verbosity(-quiet))


#################################################################################
# LOAD command
# Load JSON records into the database.
#################################################################################
@command_base.command(
    name="load", help="Load JSON records into the Electrolyte Database"
)
@click.option(
    "-f",
    "--file",
    "input_file",
    help="File to load",
    type=click.File("r"),
    default=None,
)
@click.option(
    "-t",
    "--type",
    "data_type",
    help="Type of records",
    type=click.Choice(["component", "reaction", "base"], case_sensitive=False),
    default=None,
)
@click.option(
    "-u", "--url", help="Database connection URL", default=ElectrolyteDB.DEFAULT_URL
)
@click.option(
    "-d", "--database", help="Database name", default=ElectrolyteDB.DEFAULT_DB
)
@click.option(
    "--validate/--no-validate",
    " /-n",
    help="Turn on or off validation of input",
    default=True,
)
@click.option(
    "-b",
    "--bootstrap",
    help="Bootstrap a new database by loading in the standard base data",
    is_flag=True,
    default=False,
)
def load_data(input_file, data_type, url, database, validate, bootstrap):
    try:
        edb = ElectrolyteDB(url, database)
    except ConnectionFailure as err:
        click.echo(f"Database connection failure: {err}")
        return -1
    if bootstrap:
        if not edb.is_empty():
            click.echo(
                f"Cannot bootstrap: database {database} at {url} already exists "
                f"and has one or more of the EDB collections"
            )
            return -1
        _load_bootstrap(edb, do_validate=validate)
    else:
        if input_file is None:
            click.echo("Error: -f/--file is required")
            return -1
        if data_type is None:
            click.echo("Error: -t/--type is required")
            return -2
        _load(input_file, data_type, edb, do_validate=validate)


def _load(input_file, data_type, edb, do_validate=True):
    print_messages = _log.isEnabledFor(logging.ERROR)
    filename = input_file.name
    _log.debug(f"Reading records from input file '{filename}'")
    input_data = json.load(input_file)
    if isinstance(input_data, dict):  # make single record into a list of length one
        input_data = [input_data]
    _log.info(f"Read {len(input_data)} records from input file '{filename}'")
    if do_validate:
        if data_type == "base":
            _log.warning("No validation for records of type 'base' (yet)")
            data = input_data
        else:
            _log.info("Validating records")
            if data_type in ("component", "reaction"):
                obj_type = data_type
            else:
                raise RuntimeError(f"Unexpected data type: {data_type}")
            data = []
            for record in input_data:
                try:
                    validate(record, obj_type=obj_type)
                except ValidationError as err:
                    click.echo(f"Validation failed: {err}")
                    if print_messages:
                        click.echo("Record:")
                        click.echo(json.dumps(record, indent=2))
                    return -1
                data.append(record)
    else:
        data = input_data
    _log.info(f"Loading records into collection '{data_type}'")
    n = edb.load(data, rec_type=data_type)
    if print_messages:
        click.echo(f"Loaded {n} record(s) into collection '{data_type}'")


def _load_bootstrap(edb, **kwargs):
    _log.info(f"Begin: Bootstrapping database {edb.database} at {edb.url}")
    for t in "base", "component", "reaction":
        _log.info(f"Loading collection: {t}")
        filename = t + ".json"
        path = get_edb_data(filename)
        _load(path.open("r"), t, edb, **kwargs)
    _log.info(f"End: Bootstrapping database {edb.database} at {edb.url}")


#################################################################################
# DUMP command
# Save JSON records to a file
#################################################################################


@command_base.command(
    name="dump", help="Dump JSON records from the Electrolyte Database"
)
@click.option(
    "-f",
    "--file",
    "output_file",
    required=True,
    help="File to create (will overwrite existing files!)",
    type=click.File("w"),
)
@click.option(
    "-t",
    "--type",
    "data_type",
    required=True,
    help="Type of records (MongoDB collection name)",
    type=click.Choice(["component", "reaction", "base"], case_sensitive=False),
    default=None,
)
@click.option(
    "-u", "--url", help="Database connection URL", default=ElectrolyteDB.DEFAULT_URL
)
@click.option(
    "-d", "--database", help="Database name", default=ElectrolyteDB.DEFAULT_DB
)
def dump_data(output_file, data_type, url, database):
    print_messages = _log.isEnabledFor(logging.ERROR)
    filename = output_file.name
    _log.debug(f"Writing records to output file '{filename}'")

    _log.info(f"Connecting to MongoDB at: {url}/{database}")
    db = ElectrolyteDB(url=url, db=database)

    _log.debug("Retrieving records")
    if data_type == "component":
        records = db.get_components()
    elif data_type == "reaction":
        records = db.get_reactions()
    elif data_type == "base":
        records = db.get_base()
    else:
        raise RuntimeError(f"Unexpected data type: {data_type}")

    record_list = [r.json_data for r in records]
    n = len(record_list)
    json.dump(record_list, output_file)
    if print_messages:
        click.echo(
            f"Wrote {n} record(s) from collection '{data_type}' to file '{filename}'"
        )


#################################################################################
# DROP command
# Drop a database
#################################################################################


def abort_drop_db(ctx, param, value):
    if not value:
        ctx.abort()


@command_base.command(name="drop", help="Drop a database in the Electrolyte Database")
@click.option(
    "-u", "--url", help="Database connection URL", default=ElectrolyteDB.DEFAULT_URL
)
@click.option(
    "-d", "--database", help="Database name", default=ElectrolyteDB.DEFAULT_DB
)
@click.option(
    "--yes",
    is_flag=True,
)
def drop_database(url, database, yes):
    print_messages = _log.isEnabledFor(logging.ERROR)

    # attempt to connect
    if not ElectrolyteDB.can_connect(url):
        raise click.Abort()

    if not yes:
        confirm = click.prompt(
            f"Are you sure you want to drop the database {database} at {url}",
            type=click.Choice(("y", "N"), case_sensitive=False),
            default="n",
        )
        if confirm.lower() != "y":
            raise click.Abort()

    click.echo(f"Dropping database {database} at {url} ...")
    ElectrolyteDB.drop_database(url, database)
    click.echo(f"Done")


#################################################################################
# SCHEMA command
#################################################################################


@command_base.command(name="schema", help="Show JSON schemas, in raw or readable forms")
@click.option(
    "-f",
    "--file",
    "output_file",
    help="Write output to this file instead of printing to the screen",
    type=click.File("w"),
)
@click.option(
    "-o",
    "--format",
    "output_format",
    help="Output format",
    default="markdown",
    type=click.Choice(
        ["json", "json-compact", "markdown", "html", "html-js"], case_sensitive=False
    ),
)
@click.option(
    "-t",
    "--type",
    "data_type",
    required=True,
    help="Type of records",
    type=click.Choice(["component", "reaction"], case_sensitive=False),
    default=None,
)
@click.option(
    "-u", "--url", help="Database connection URL", default=ElectrolyteDB.DEFAULT_URL
)
@click.option(
    "-d", "--database", help="Database name", default=ElectrolyteDB.DEFAULT_DB
)
def schema(output_file, output_format, data_type, url, database):
    print_messages = _log.isEnabledFor(logging.ERROR)
    if output_file:
        stream = output_file
    else:
        stream = sys.stdout
    schema_data = edb_schemas[data_type]
    if output_format == "json":
        json.dump(schema_data, stream, indent=2)
    else:
        if output_format == "markdown":
            tmpl = "md"
        elif output_format == "html":
            tmpl = "flat"
        elif output_format == "html-js":
            tmpl = "js"
        config = GenerationConfiguration(template_name=tmpl)
        with tempfile.TemporaryDirectory() as tmpdir_name:
            schema_path = pathlib.Path(tmpdir_name) / "schema.json"
            with schema_path.open("w+", encoding="utf-8") as schema_file:
                json.dump(schema_data, schema_file)
                schema_file.seek(0)
                schema_gen.generate_from_file_object(schema_file, stream, config=config)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains utility functions for initialization of WaterTAP models.
"""


__author__ = "Adam Atia"

from pyomo.environ import check_optimal_termination, Var
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.scaling import get_scaling_factor, __none_left_mult
from idaes.core.util import get_solver
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)

def check_solve(results, checkpoint=None, logger=_log, fail_flag=False):
    """
    Check that solver termination is optimal and OK in an initialization routine.
    If the check fails, proceed through initialization with only a logger warning by default,
    or set fail_flag=True to raise an error. This should also work for checking a solve outside
    of an initialization routine.

    Keyword Arguments:
            results : solver results
            checkpoint : Optional string argument to specify the step of initialization being checked
                        (e.g., checkpoint="Initialization step 1: solve indexed blocks")
            logger : Optional argument for loading idaes.getInitLogger object (e.g., logger=init_log)
            fail_flag : Boolean argument to specify error or warning (Default: fail_flag=False produces logger warning.
                        set fail_flag=True to raise an error and stop the initialization routine.)
    Returns:
        None

    """
    if check_optimal_termination(results):
        if checkpoint is None:
            logger.info(f'Solve successful.')
        else:
            logger.info(f'{checkpoint} successful.')
    else:
        if checkpoint is None:
            msg = f"The solver failed to converge to an optimal solution. " \
                  f"This suggests that the user provided infeasible inputs or that the model is poorly scaled."
        else:
            msg = f"{checkpoint} failed. The solver failed to converge to an optimal solution. " \
                  f"This suggests that the user provided infeasible inputs or that the model is poorly scaled."
        if fail_flag:
            logger.error(msg)
            raise ValueError(msg)
        else:
            logger.warning(msg)


def check_dof(blk, fail_flag=False, logger=_log, expected_dof=0):
    """
    Check that degrees of freedom are 0, or the expected amount ``expected_dof``.
    If not 0 or ``expected_dof``, either throw a warning and continue or throw an error and stop.

    Keyword Arguments:
            blk : block to check
            fail_flag : Boolean argument to specify error or warning
            (Default: fail_flag=False produces logger warning. Set fail_flag=True to raise an error and stop
             the initialization routine.)
            logger : Optional argument for loading idaes.getInitLogger object (e.g., logger=init_log)
            expected_dof : Integer number of degrees of freedom ``blk`` should have

    Returns:
        None

    """
    if degrees_of_freedom(blk) != expected_dof:
        if expected_dof == 0:
            msg = f"Non-zero degrees of freedom: Degrees of freedom on {blk} = {degrees_of_freedom(blk)}. " \
                  f"Fix {degrees_of_freedom(blk)} more variable(s)"
        elif degrees_of_freedom(blk) < expected_dof:
            msg = f"Unexpected degrees of freedom: Degrees of freedom on {blk} = {degrees_of_freedom(blk)}. " \
                  f"Expected {expected_dof}. Unfix {expected_dof - degrees_of_freedom(blk)} variable(s)"
        elif degrees_of_freedom(blk) > expected_dof:
            msg = f"Unexpected degrees of freedom: Degrees of freedom on {blk} = {degrees_of_freedom(blk)}. " \
                  f"Expected {expected_dof}. Fix {degrees_of_freedom(blk) - expected_dof} variable(s)"
        if fail_flag:
            logger.error(msg)
            raise ValueError(msg)
        else:
            logger.warning(msg)


def assert_degrees_of_freedom(blk, expected_dof):
    """
    Assert that degrees of freedom are ``expected_dof``.
    If not ``expected_dof``, throw an error and stop.

    Keyword Arguments:
            blk : block to check
            expected_dof : Integer number of degrees of freedom ``blk`` should have

    Returns:
        None

    """
    check_dof(blk, True, expected_dof=expected_dof)


def assert_no_degrees_of_freedom(blk):
    """
    Assert that degrees of freedom are 0.
    If ``blk`` has non-zero degrees of freedom, throw an error and stop.

    Keyword Arguments:
            blk : block to check

    Returns:
        None

    """
    check_dof(blk, True)


def assert_no_initialization_perturbation(blk, optarg=None, solver=None):
    """
    Assert that IPOPT will *not* move the initialization

    Args:
        blk: Pyomo block
        optarg: IPOPT options (default=None) (Should be None if solver is specified)
        solver: Pyomo IPOPT solver instance (default=None) (Should be None if optarg is specified).

    Returns:
        None
    """
    if solver is not None and optarg is not None:
        raise ValueError("Supply a solver or optarg, not both")
    if optarg is None:
        optarg = {}
    if solver is None:
        solver = get_solver(options=optarg)
    if solver.name not in ('ipopt',
            'ipopt-watertap',
            ):
        raise ValueError(f"Solver {solver.name} is not supported")

    options = solver.options
    bound_push = options.get('bound_push', 1e-2)
    bound_frac = options.get('bound_frac', 1e-2)
    bound_relax_factor = options.get('bound_relax_factor', 1e-8)
    if solver.name == 'ipopt-watertap':
        user_scaling = True
    else:
        user_scaling = (options.get('nlp_scaling_method', 'gradient-based') == 'user-scaling')

    for v, val, result in generate_initialization_perturbation(blk, bound_push, bound_frac, bound_relax_factor, user_scaling):
        raise ValueError(f"IPOPT will move scaled initial value for variable {v.name} from {val:e} to {result:e}")


def print_initialization_perturbation(blk, bound_push=1e-2, bound_frac=1e-2, bound_relax_factor=1e-8, user_scaling=False):
    """
    Print the initialization perturbations performed by IPOPT for a given Block

    Args:
        blk: Pyomo block
        bound_push: bound_push to evaluate (same as IPOPT option) (default=1e-2)
        bound_frac: bound_frac to evaluate (same as IPOPT option) (default=1e-2)
        bound_relax_factor: bound_relax_factor to evaluate (same as IPOPT option) (default=1e-8)
        user_scaling: If True, the variables are scaled as if `nlp_scaling_method = user-scaling`
                       is used. (default=False)

    Returns:
        None
    """
    for v, val, result in generate_initialization_perturbation(blk, bound_push, bound_frac, bound_relax_factor, user_scaling):
        print(f"IPOPT will move scaled initial value for variable {v.name} from {val:e} to {result:e}")


def generate_initialization_perturbation(blk, bound_push=1e-2, bound_frac=1e-2, bound_relax_factor=1e-8, user_scaling=False):
    """
    Generate the initialization perturbations performed by IPOPT for a given Block

    Args:
        blk: Pyomo block
        bound_push: bound_push to evaluate (same as IPOPT option) (default=1e-2)
        bound_frac: bound_frac to evaluate (same as IPOPT option) (default=1e-2)
        bound_relax_factor: bound_relax_factor to evaluate (same as IPOPT option) (default=1e-8)
        user_scaling: If True, the variables are scaled as if `nlp_scaling_method = user-scaling`
                       is used. (default=False)

    Yields:
        tuple: (pyo.Var object, current_value, perturbed_value)
    """
    kappa1 = bound_push; kappa2 = bound_frac
    for v in blk.component_data_objects(Var):
        if v.value is None:
            _log.warning(f"Variable {v.name} has no initial value")
            continue
        if v.fixed:
            continue
        if user_scaling:
            sf = get_scaling_factor(v, default=1.)
        else:
            sf = 1.
        v_lb = __none_left_mult(v.lb,sf)
        if v_lb is not None:
            v_lb -= bound_relax_factor*max(1, abs(v_lb))
        v_value = v.value*sf
        v_ub = __none_left_mult(v.ub,sf)
        if v_ub is not None:
            v_ub += bound_relax_factor*max(1, abs(v_ub))
        if v_lb is not None:
            if v_ub is not None:
                pl = min( kappa1*max(1, abs(v_lb)), kappa2*(v_ub - v_lb) )
            else:
                pl = kappa1*max(1, abs(v_lb))
            if v_value < v_lb + pl:
                yield (v, v_value, v_lb+pl)
        if v_ub is not None:
            if v_lb is not None:
                pu = min( kappa1*max(1, abs(v_ub)), kappa2*(v_ub - v_lb) )
            else:
                pu = kappa1*max(1, abs(v_ub))
            if v_value > v_ub - pu:
                yield (v, v_value, v_ub-pu)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

import pytest
from pyomo.environ import (ConcreteModel,
                           value,
                           Param,
                           Var,
                           Expression,
                           Constraint,
                           assert_optimal_termination)
from pyomo.network import Port
from idaes.core import (FlowsheetBlock,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType,
                        ControlVolume0DBlock)
from watertap.unit_models.reverse_osmosis_0D import (ReverseOsmosis0D,
                                                       ConcentrationPolarizationType,
                                                       MassTransferCoefficient,
                                                       PressureChangeType)
import watertap.property_models.NaCl_prop_pack\
    as props

from idaes.core.util import get_solver
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              number_unused_variables)
from idaes.core.util.testing import initialization_tester
from idaes.core.util.scaling import (calculate_scaling_factors,
                                     unscaled_variables_generator,
                                     unscaled_constraints_generator,
                                     badly_scaled_var_generator)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = props.NaClParameterBlock()
    m.fs.unit = ReverseOsmosis0D(default={"property_package": m.fs.properties})

    assert len(m.fs.unit.config) == 11

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.material_balance_type == \
           MaterialBalanceType.useDefault
    assert m.fs.unit.config.energy_balance_type == \
           EnergyBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == \
           MomentumBalanceType.pressureTotal
    assert not m.fs.unit.config.has_pressure_change
    assert m.fs.unit.config.property_package is \
           m.fs.properties
    assert m.fs.unit.config.concentration_polarization_type == \
           ConcentrationPolarizationType.calculated
    assert m.fs.unit.config.mass_transfer_coefficient == \
           MassTransferCoefficient.calculated
    assert m.fs.unit.config.pressure_change_type == \
           PressureChangeType.fixed_per_stage

@pytest.mark.unit
def test_option_has_pressure_change():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = props.NaClParameterBlock()
    m.fs.unit = ReverseOsmosis0D(default={
        "property_package": m.fs.properties,
        "has_pressure_change": True})

    assert isinstance(m.fs.unit.feed_side.deltaP, Var)
    assert isinstance(m.fs.unit.deltaP, Var)

@pytest.mark.unit
def test_option_concentration_polarization_type_fixed():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = props.NaClParameterBlock()
    m.fs.unit = ReverseOsmosis0D(default={
        "property_package": m.fs.properties,
        "has_pressure_change": True,
        "concentration_polarization_type": ConcentrationPolarizationType.fixed,
        "mass_transfer_coefficient": MassTransferCoefficient.none})

    assert m.fs.unit.config.concentration_polarization_type == \
           ConcentrationPolarizationType.fixed
    assert isinstance(m.fs.unit.cp_modulus, Var)

@pytest.mark.unit
def test_option_concentration_polarization_type_calculated_kf_fixed():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = props.NaClParameterBlock()
    m.fs.unit = ReverseOsmosis0D(default={
        "property_package": m.fs.properties,
        "has_pressure_change": True,
        "concentration_polarization_type": ConcentrationPolarizationType.calculated,
        "mass_transfer_coefficient": MassTransferCoefficient.fixed})

    assert m.fs.unit.config.concentration_polarization_type == \
           ConcentrationPolarizationType.calculated
    assert m.fs.unit.config.mass_transfer_coefficient == \
           MassTransferCoefficient.fixed
    assert isinstance(m.fs.unit.Kf_io, Var)

@pytest.mark.unit
def test_option_concentration_polarization_type_calculated_kf_calculated():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = props.NaClParameterBlock()
    m.fs.unit = ReverseOsmosis0D(default={
        "property_package": m.fs.properties,
        "has_pressure_change": True,
        "concentration_polarization_type": ConcentrationPolarizationType.calculated,
        "mass_transfer_coefficient": MassTransferCoefficient.calculated})

    assert m.fs.unit.config.concentration_polarization_type == \
           ConcentrationPolarizationType.calculated
    assert m.fs.unit.config.mass_transfer_coefficient == \
           MassTransferCoefficient.calculated
    assert isinstance(m.fs.unit.Kf_io, Var)
    assert isinstance(m.fs.unit.channel_height, Var)
    assert isinstance(m.fs.unit.width, Var)
    assert isinstance(m.fs.unit.length, Var)
    assert isinstance(m.fs.unit.dh, Var)
    assert isinstance(m.fs.unit.spacer_porosity, Var)
    assert isinstance(m.fs.unit.N_Sc_io, Var)
    assert isinstance(m.fs.unit.N_Sh_io, Var)
    assert isinstance(m.fs.unit.N_Re_io, Var)

@pytest.mark.unit
def test_option_pressure_change_calculated():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = props.NaClParameterBlock()
    m.fs.unit = ReverseOsmosis0D(default={
        "property_package": m.fs.properties,
        "has_pressure_change": True,
        "concentration_polarization_type": ConcentrationPolarizationType.none,
        "mass_transfer_coefficient": MassTransferCoefficient.none,
        "pressure_change_type": PressureChangeType.calculated})

    assert m.fs.unit.config.concentration_polarization_type == \
           ConcentrationPolarizationType.none
    assert m.fs.unit.config.mass_transfer_coefficient == \
           MassTransferCoefficient.none
    assert m.fs.unit.config.pressure_change_type == \
           PressureChangeType.calculated
    assert isinstance(m.fs.unit.feed_side.deltaP, Var)
    assert isinstance(m.fs.unit.deltaP, Var)
    assert isinstance(m.fs.unit.channel_height, Var)
    assert isinstance(m.fs.unit.width, Var)
    assert isinstance(m.fs.unit.length, Var)
    assert isinstance(m.fs.unit.dh, Var)
    assert isinstance(m.fs.unit.spacer_porosity, Var)
    assert isinstance(m.fs.unit.N_Re_io, Var)


class TestReverseOsmosis():
    @pytest.fixture(scope="class")
    def RO_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = props.NaClParameterBlock()

        m.fs.unit = ReverseOsmosis0D(default={
            "property_package": m.fs.properties,
            "has_pressure_change": True,
            "concentration_polarization_type": ConcentrationPolarizationType.fixed,
            "mass_transfer_coefficient": MassTransferCoefficient.none})

        # fully specify system
        feed_flow_mass = 1
        feed_mass_frac_NaCl = 0.035
        feed_pressure = 50e5
        feed_temperature = 273.15 + 25
        membrane_pressure_drop = 3e5
        membrane_area = 50
        A = 4.2e-12
        B = 3.5e-8
        pressure_atmospheric = 101325
        concentration_polarization_modulus = 1.1

        feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
        m.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'NaCl'].fix(
            feed_flow_mass * feed_mass_frac_NaCl)
        m.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'H2O'].fix(
            feed_flow_mass * feed_mass_frac_H2O)
        m.fs.unit.inlet.pressure[0].fix(feed_pressure)
        m.fs.unit.inlet.temperature[0].fix(feed_temperature)
        m.fs.unit.deltaP.fix(-membrane_pressure_drop)
        m.fs.unit.area.fix(membrane_area)
        m.fs.unit.A_comp.fix(A)
        m.fs.unit.B_comp.fix(B)
        m.fs.unit.permeate.pressure[0].fix(pressure_atmospheric)
        m.fs.unit.cp_modulus.fix(concentration_polarization_modulus)
        return m

    @pytest.mark.unit
    def test_build(self, RO_frame):
        m = RO_frame

        # test ports
        port_lst = ['inlet', 'retentate', 'permeate']
        for port_str in port_lst:
            port = getattr(m.fs.unit, port_str)
            assert len(port.vars) == 3  # number of state variables for NaCl property package
            assert isinstance(port, Port)

        # test pyomo objects on unit
        unit_objs_type_dict = {'dens_solvent': Param,
                               'A_comp': Var,
                               'B_comp': Var,
                               'flux_mass_io_phase_comp': Var,
                               'area': Var,
                               'recovery_vol_phase': Var,
                               'recovery_mass_phase_comp': Var,
                               'rejection_phase_comp': Var,
                               'deltaP': Var,
                               'cp_modulus': Var,
                               'mass_transfer_phase_comp': Var,
                               'flux_mass_phase_comp_avg': Expression,
                               'over_pressure_ratio': Expression,
                               'eq_mass_transfer_term': Constraint,
                               'eq_permeate_production': Constraint,
                               'eq_flux_io': Constraint,
                               'eq_connect_mass_transfer': Constraint,
                               'eq_connect_enthalpy_transfer': Constraint,
                               'eq_permeate_isothermal': Constraint,
                               'eq_recovery_vol_phase': Constraint,
                               'eq_recovery_mass_phase_comp': Constraint,
                               'eq_rejection_phase_comp': Constraint,
                               }
        for (obj_str, obj_type) in unit_objs_type_dict.items():
            obj = getattr(m.fs.unit, obj_str)
            assert isinstance(obj, obj_type)
        # check that all added unit objects are tested
        for obj in m.fs.unit.component_objects(
                [Param, Var, Expression, Constraint], descend_into=False):
            obj_str = obj.local_name
            if obj_str[0] == '_':
                continue  # do not test hidden references
            assert obj_str in unit_objs_type_dict

        # test feed-side control volume and associated stateblocks
        assert isinstance(m.fs.unit.feed_side, ControlVolume0DBlock)
        cv_stateblock_lst = ['properties_in', 'properties_out',
                             'properties_interface_in', 'properties_interface_out']
        for sb_str in cv_stateblock_lst:
            sb = getattr(m.fs.unit.feed_side, sb_str)
            assert isinstance(sb, props.NaClStateBlock)
        # test objects added to control volume
        cv_objs_type_dict = {'eq_concentration_polarization_io': Constraint,
                             'eq_equal_temp_interface_io': Constraint,
                             'eq_equal_pressure_interface_io': Constraint,
                             'eq_equal_flow_vol_interface_io': Constraint}
        for (obj_str, obj_type) in cv_objs_type_dict.items():
            obj = getattr(m.fs.unit.feed_side, obj_str)
            assert isinstance(obj, obj_type)

        # test permeate-side control volume and associated stateblocks
        assert isinstance(m.fs.unit.permeate_side, ControlVolume0DBlock)
        cv_stateblock_lst = ['properties_in', 'properties_out', 'properties_mixed']
        for sb_str in cv_stateblock_lst:
            sb = getattr(m.fs.unit.permeate_side, sb_str)
            assert isinstance(sb, props.NaClStateBlock)
        # test objects added to control volume
        cv_objs_type_dict = {'eq_mass_frac_permeate_io': Constraint,
                             'eq_temperature_permeate_io': Constraint,
                             'eq_pressure_permeate_io': Constraint,
                             'eq_flow_vol_permeate_io': Constraint}
        for (obj_str, obj_type) in cv_objs_type_dict.items():
            obj = getattr(m.fs.unit.permeate_side, obj_str)
            assert isinstance(obj, obj_type)

        # test statistics
        assert number_variables(m) == 124
        assert number_total_constraints(m) == 96
        assert number_unused_variables(m) == 7  # vars from property package parameters

    @pytest.mark.unit
    def test_dof(self, RO_frame):
        m = RO_frame
        assert degrees_of_freedom(m) == 0

    @pytest.mark.unit
    def test_calculate_scaling(self, RO_frame):
        m = RO_frame

        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1, index=('Liq', 'H2O'))
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'NaCl'))
        calculate_scaling_factors(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(unscaled_variables_generator(m))
        assert len(unscaled_var_list) == 0
        # check that all constraints have been scaled
        unscaled_constraint_list = list(unscaled_constraints_generator(m))
        assert len(unscaled_constraint_list) == 0

        # check repeated scaling does not create badly scaled vars
        calculate_scaling_factors(m)
        for _ in badly_scaled_var_generator(m):
            assert False

    @pytest.mark.component
    def test_initialize(self, RO_frame):
        initialization_tester(RO_frame, fail_on_warning=True)

    @pytest.mark.component
    def test_var_scaling(self, RO_frame):
        m = RO_frame
        badly_scaled_var_lst = list(badly_scaled_var_generator(m))
        assert badly_scaled_var_lst == []

    @pytest.mark.component
    def test_solve(self, RO_frame):
        m = RO_frame
        results = solver.solve(m)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_conservation(self, RO_frame):
        m = RO_frame
        b = m.fs.unit
        comp_lst = ['NaCl', 'H2O']

        flow_mass_inlet = sum(
            b.feed_side.properties_in[0].flow_mass_phase_comp['Liq', j] for j in comp_lst)
        flow_mass_retentate = sum(
            b.feed_side.properties_out[0].flow_mass_phase_comp['Liq', j] for j in comp_lst)
        flow_mass_permeate = sum(
            b.permeate_side.properties_mixed[0].flow_mass_phase_comp['Liq', j] for j in comp_lst)

        assert (abs(value(flow_mass_inlet - flow_mass_retentate - flow_mass_permeate
                          )) <= 1e-5)

        assert (abs(value(
            flow_mass_inlet * b.feed_side.properties_in[0].enth_mass_phase['Liq']
            - flow_mass_retentate * b.feed_side.properties_out[0].enth_mass_phase['Liq']
            - flow_mass_permeate * b.permeate_side.properties_mixed[0].enth_mass_phase['Liq']
        )) <= 1e-5)

    @pytest.mark.component
    def test_solution(self, RO_frame):
        m = RO_frame
        assert (pytest.approx(4.722e-3, rel=1e-3) ==
                value(m.fs.unit.flux_mass_phase_comp_avg[0, 'Liq', 'H2O']))
        assert (pytest.approx(1.576e-6, rel=1e-3) ==
                value(m.fs.unit.flux_mass_phase_comp_avg[0, 'Liq', 'NaCl']))
        assert (pytest.approx(0.2361, rel=1e-3) ==
                value(m.fs.unit.permeate_side.properties_mixed[0].flow_mass_phase_comp['Liq', 'H2O']))
        assert (pytest.approx(7.879e-5, rel=1e-3) ==
                value(m.fs.unit.permeate_side.properties_mixed[0].flow_mass_phase_comp['Liq', 'NaCl']))
        assert (pytest.approx(value(m.fs.unit.cp_modulus[0, 'NaCl']), rel=1e-3) ==
                value(m.fs.unit.feed_side.properties_interface_in[0].conc_mass_phase_comp['Liq', 'NaCl'])
                / value(m.fs.unit.feed_side.properties_in[0].conc_mass_phase_comp['Liq', 'NaCl']))
        assert (pytest.approx(value(m.fs.unit.cp_modulus[0, 'NaCl']), rel=1e-3) ==
                value(m.fs.unit.feed_side.properties_interface_out[0].conc_mass_phase_comp['Liq', 'NaCl'])
                / value(m.fs.unit.feed_side.properties_out[0].conc_mass_phase_comp['Liq', 'NaCl']))
        assert (pytest.approx(-3.000e5, rel=1e-3) == value(m.fs.unit.deltaP[0]))

    @pytest.mark.component
    def test_CP_calculation_with_kf_fixed(self):
        """ Testing 0D-RO with ConcentrationPolarizationType.calculated option enabled.
        This option makes use of an alternative constraint for the feed-side, membrane-interface concentration.
        Additionally, two more variables are created when this option is enabled: Kf_io - feed-channel
        mass transfer coefficients at the channel inlet and outlet.
        """
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = props.NaClParameterBlock()

        m.fs.unit = ReverseOsmosis0D(default={
            "property_package": m.fs.properties,
            "has_pressure_change": True,
            "concentration_polarization_type": ConcentrationPolarizationType.calculated,
            "mass_transfer_coefficient": MassTransferCoefficient.fixed})

        # fully specify system
        feed_flow_mass = 1
        feed_mass_frac_NaCl = 0.035
        feed_pressure = 50e5
        feed_temperature = 273.15 + 25
        membrane_pressure_drop = 3e5
        membrane_area = 50
        A = 4.2e-12
        B = 3.5e-8
        pressure_atmospheric = 101325
        kf = 2e-5

        feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
        m.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'NaCl'].fix(
            feed_flow_mass * feed_mass_frac_NaCl)
        m.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'H2O'].fix(
            feed_flow_mass * feed_mass_frac_H2O)
        m.fs.unit.inlet.pressure[0].fix(feed_pressure)
        m.fs.unit.inlet.temperature[0].fix(feed_temperature)
        m.fs.unit.deltaP.fix(-membrane_pressure_drop)
        m.fs.unit.area.fix(membrane_area)
        m.fs.unit.A_comp.fix(A)
        m.fs.unit.B_comp.fix(B)
        m.fs.unit.permeate.pressure[0].fix(pressure_atmospheric)
        m.fs.unit.Kf_io[0, 'in', 'NaCl'].fix(kf)
        m.fs.unit.Kf_io[0, 'out', 'NaCl'].fix(kf)

        # test statistics
        assert number_variables(m) == 125
        assert number_total_constraints(m) == 96
        assert number_unused_variables(m) == 7  # vars from property package parameters

        # test degrees of freedom
        assert degrees_of_freedom(m) == 0

        # test scaling
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1, index=('Liq', 'H2O'))
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'NaCl'))
        calculate_scaling_factors(m)

        # check that all variables have scaling factors.
        # TODO: Setting the "include_fixed" arg as True reveals
        #  unscaled vars that weren't being accounted for previously. However, calling the whole block (i.e.,
        #  m) shows that several NaCl property parameters are unscaled. For now, we are just interested in ensuring
        #  unit variables are scaled (hence, calling m.fs.unit) but might need to revisit scaling and associated
        #  testing for property models.

        unscaled_var_list = list(unscaled_variables_generator(m.fs.unit, include_fixed=True))
        assert len(unscaled_var_list) == 0

        # check that all constraints have been scaled
        unscaled_constraint_list = list(unscaled_constraints_generator(m))
        assert len(unscaled_constraint_list) == 0

        # # test initialization
        initialization_tester(m, fail_on_warning=True)

        # test variable scaling
        badly_scaled_var_lst = list(badly_scaled_var_generator(m))
        assert badly_scaled_var_lst == []

        # test solve
        results = solver.solve(m)

        # Check for optimal solution
        assert_optimal_termination(results)

        # test solution
        assert (pytest.approx(3.815e-3, rel=1e-3) ==
                value(m.fs.unit.flux_mass_phase_comp_avg[0, 'Liq', 'H2O']))
        assert (pytest.approx(1.668e-6, rel=1e-3) ==
                value(m.fs.unit.flux_mass_phase_comp_avg[0, 'Liq', 'NaCl']))
        assert (pytest.approx(0.1908, rel=1e-3) ==
                value(m.fs.unit.permeate_side.properties_mixed[0].flow_mass_phase_comp['Liq', 'H2O']))
        assert (pytest.approx(8.337e-5, rel=1e-3) ==
                value(m.fs.unit.permeate_side.properties_mixed[0].flow_mass_phase_comp['Liq', 'NaCl']))
        assert (pytest.approx(46.07, rel=1e-3) ==
                value(m.fs.unit.feed_side.properties_interface_in[0].conc_mass_phase_comp['Liq', 'NaCl']))
        assert (pytest.approx(44.34, rel=1e-3) ==
                value(m.fs.unit.feed_side.properties_out[0].conc_mass_phase_comp['Liq', 'NaCl']))
        assert (pytest.approx(50.20, rel=1e-3) ==
                value(m.fs.unit.feed_side.properties_interface_out[0].conc_mass_phase_comp['Liq', 'NaCl']))

    @pytest.mark.component
    def test_CP_calculation_with_kf_calculation(self):
        """ Testing 0D-RO with ConcentrationPolarizationType.calculated option and MassTransferCoefficient.calculated
        option enabled.
        """
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = props.NaClParameterBlock()

        m.fs.unit = ReverseOsmosis0D(default={
            "property_package": m.fs.properties,
            "has_pressure_change": True,
            "concentration_polarization_type": ConcentrationPolarizationType.calculated,
            "mass_transfer_coefficient": MassTransferCoefficient.calculated})

        # fully specify system
        feed_flow_mass = 1
        feed_mass_frac_NaCl = 0.035
        feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
        feed_pressure = 50e5
        feed_temperature = 273.15 + 25
        membrane_pressure_drop = 3e5
        length = 20
        membrane_area = 50
        A = 4.2e-12
        B = 3.5e-8
        pressure_atmospheric = 101325

        m.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'NaCl'].fix(
            feed_flow_mass * feed_mass_frac_NaCl)
        m.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'H2O'].fix(
            feed_flow_mass * feed_mass_frac_H2O)
        m.fs.unit.inlet.pressure[0].fix(feed_pressure)
        m.fs.unit.inlet.temperature[0].fix(feed_temperature)
        m.fs.unit.deltaP.fix(-membrane_pressure_drop)
        m.fs.unit.area.fix(membrane_area)
        m.fs.unit.A_comp.fix(A)
        m.fs.unit.B_comp.fix(B)
        m.fs.unit.permeate.pressure[0].fix(pressure_atmospheric)

        m.fs.unit.channel_height.fix(0.002)
        m.fs.unit.spacer_porosity.fix(0.75)
        m.fs.unit.length.fix(length)

        # test statistics
        assert number_variables(m) == 140
        assert number_total_constraints(m) == 110
        assert number_unused_variables(m) == 0  # vars from property package parameters

        # test degrees of freedom
        assert degrees_of_freedom(m) == 0

        # test scaling
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1, index=('Liq', 'H2O'))
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'NaCl'))

        calculate_scaling_factors(m)

        # check that all variables have scaling factors.
        # TODO: see aforementioned TODO on revisiting scaling and associated testing for property models.
        unscaled_var_list = list(unscaled_variables_generator(m.fs.unit, include_fixed=True))
        assert len(unscaled_var_list) == 0

        # check that all constraints have been scaled
        unscaled_constraint_list = list(unscaled_constraints_generator(m))
        assert len(unscaled_constraint_list) == 0

        # test initialization
        initialization_tester(m, fail_on_warning=True)

        # test variable scaling
        badly_scaled_var_lst = list(badly_scaled_var_generator(m))
        assert badly_scaled_var_lst == []

        # test solve
        results = solver.solve(m, tee=True)

        # Check for optimal solution
        assert_optimal_termination(results)

        # test solution
        assert (pytest.approx(4.562e-3, rel=1e-3) ==
                value(m.fs.unit.flux_mass_phase_comp_avg[0, 'Liq', 'H2O']))
        assert (pytest.approx(1.593e-6, rel=1e-3) ==
                value(m.fs.unit.flux_mass_phase_comp_avg[0, 'Liq', 'NaCl']))
        assert (pytest.approx(0.2281, rel=1e-3) ==
                value(m.fs.unit.permeate_side.properties_mixed[0].flow_mass_phase_comp['Liq', 'H2O']))
        assert (pytest.approx(7.963e-5, rel=1e-3) ==
                value(m.fs.unit.permeate_side.properties_mixed[0].flow_mass_phase_comp['Liq', 'NaCl']))
        assert (pytest.approx(35.75, rel=1e-3) ==
                value(m.fs.unit.feed_side.properties_in[0].conc_mass_phase_comp['Liq', 'NaCl']))
        assert (pytest.approx(41.96, rel=1e-3) ==
                value(m.fs.unit.feed_side.properties_interface_in[0].conc_mass_phase_comp['Liq', 'NaCl']))
        assert (pytest.approx(46.57, rel=1e-3) ==
                value(m.fs.unit.feed_side.properties_out[0].conc_mass_phase_comp['Liq', 'NaCl']))
        assert (pytest.approx(49.94, rel=1e-3) ==
                value(m.fs.unit.feed_side.properties_interface_out[0].conc_mass_phase_comp['Liq', 'NaCl']))

    @pytest.mark.component
    def test_Pdrop_calculation(self):
        """ Testing 0D-RO with PressureChangeType.calculated option.
        """
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = props.NaClParameterBlock()

        m.fs.unit = ReverseOsmosis0D(default={
            "property_package": m.fs.properties,
            "has_pressure_change": True,
            "concentration_polarization_type": ConcentrationPolarizationType.calculated,
            "mass_transfer_coefficient": MassTransferCoefficient.calculated,
            "pressure_change_type": PressureChangeType.calculated})

        # fully specify system
        feed_flow_mass = 1/3.6
        feed_mass_frac_NaCl = 0.035
        feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
        feed_pressure = 70e5
        feed_temperature = 273.15 + 25
        membrane_area = 19
        A = 4.2e-12
        B = 3.5e-8
        pressure_atmospheric = 101325

        m.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'NaCl'].fix(
            feed_flow_mass * feed_mass_frac_NaCl)
        m.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'H2O'].fix(
            feed_flow_mass * feed_mass_frac_H2O)
        m.fs.unit.inlet.pressure[0].fix(feed_pressure)
        m.fs.unit.inlet.temperature[0].fix(feed_temperature)
        m.fs.unit.area.fix(membrane_area)
        m.fs.unit.A_comp.fix(A)
        m.fs.unit.B_comp.fix(B)
        m.fs.unit.permeate.pressure[0].fix(pressure_atmospheric)
        m.fs.unit.channel_height.fix(0.001)
        m.fs.unit.spacer_porosity.fix(0.97)
        m.fs.unit.length.fix(16)

        # test statistics
        assert number_variables(m) == 146
        assert number_total_constraints(m) == 117
        assert number_unused_variables(m) == 0  # vars from property package parameters

        # test degrees of freedom
        assert degrees_of_freedom(m) == 0

        # test scaling
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1, index=('Liq', 'H2O'))
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'NaCl'))

        calculate_scaling_factors(m)

        # check that all variables have scaling factors.
        # TODO: see aforementioned TODO on revisiting scaling and associated testing for property models.
        unscaled_var_list = list(unscaled_variables_generator(m.fs.unit, include_fixed=True))
        assert len(unscaled_var_list) == 0

        # check that all constraints have been scaled
        unscaled_constraint_list = list(unscaled_constraints_generator(m))
        assert len(unscaled_constraint_list) == 0

        # test initialization
        initialization_tester(m, fail_on_warning=True)

        # test variable scaling
        badly_scaled_var_lst = list(badly_scaled_var_generator(m))
        assert badly_scaled_var_lst == []

        # test solve
        results = solver.solve(m, tee=True)

        # Check for optimal solution
        assert_optimal_termination(results)

        # test solution
        assert (pytest.approx(-1.661e5, rel=1e-3) == value(m.fs.unit.deltaP[0]))
        assert (pytest.approx(-1.038e4, rel=1e-3) == value(m.fs.unit.deltaP[0]/m.fs.unit.length))
        assert (pytest.approx(395.8, rel=1e-3) == value(m.fs.unit.N_Re_io[0, 'in']))
        assert (pytest.approx(0.2361, rel=1e-3) == value(m.fs.unit.velocity_io[0, 'in']))
        assert (pytest.approx(191.1, rel=1e-3) == value(m.fs.unit.N_Re_io[0, 'out']))
        assert (pytest.approx(0.1187, rel=1e-3) == value(m.fs.unit.velocity_io[0, 'out']))
        assert (pytest.approx(7.089e-3, rel=1e-3) ==
                value(m.fs.unit.flux_mass_phase_comp_avg[0, 'Liq', 'H2O']))
        assert (pytest.approx(2.188e-6, rel=1e-3) ==
                value(m.fs.unit.flux_mass_phase_comp_avg[0, 'Liq', 'NaCl']))
        assert (pytest.approx(0.1347, rel=1e-3) ==
                value(m.fs.unit.permeate_side.properties_mixed[0].flow_mass_phase_comp['Liq', 'H2O']))
        assert (pytest.approx(4.157e-5, rel=1e-3) ==
                value(m.fs.unit.permeate_side.properties_mixed[0].flow_mass_phase_comp['Liq', 'NaCl']))
        assert (pytest.approx(50.08, rel=1e-3) ==
                value(m.fs.unit.feed_side.properties_interface_in[0].conc_mass_phase_comp['Liq', 'NaCl']))
        assert (pytest.approx(70.80, rel=1e-3) ==
                value(m.fs.unit.feed_side.properties_out[0].conc_mass_phase_comp['Liq', 'NaCl']))
        assert (pytest.approx(76.32, rel=1e-3) ==
                value(m.fs.unit.feed_side.properties_interface_out[0].conc_mass_phase_comp['Liq', 'NaCl']))

    @pytest.mark.component
    def test_Pdrop_fixed_per_unit_length(self):
        """ Testing 0D-RO with PressureChangeType.fixed_per_unit_length option.
        """
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = props.NaClParameterBlock()

        m.fs.unit = ReverseOsmosis0D(default={
            "property_package": m.fs.properties,
            "has_pressure_change": True,
            "concentration_polarization_type": ConcentrationPolarizationType.calculated,
            "mass_transfer_coefficient": MassTransferCoefficient.calculated,
            "pressure_change_type": PressureChangeType.fixed_per_unit_length})

        # fully specify system
        feed_flow_mass = 1
        feed_mass_frac_NaCl = 0.035
        feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
        feed_pressure = 50e5
        feed_temperature = 273.15 + 25
        membrane_area = 50
        length = 20
        A = 4.2e-12
        B = 3.5e-8
        pressure_atmospheric = 101325
        membrane_pressure_drop = 3e5

        m.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'NaCl'].fix(
            feed_flow_mass * feed_mass_frac_NaCl)
        m.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'H2O'].fix(
            feed_flow_mass * feed_mass_frac_H2O)
        m.fs.unit.inlet.pressure[0].fix(feed_pressure)
        m.fs.unit.inlet.temperature[0].fix(feed_temperature)
        m.fs.unit.area.fix(membrane_area)
        m.fs.unit.A_comp.fix(A)
        m.fs.unit.B_comp.fix(B)
        m.fs.unit.permeate.pressure[0].fix(pressure_atmospheric)

        m.fs.unit.channel_height.fix(0.002)
        m.fs.unit.spacer_porosity.fix(0.75)
        m.fs.unit.length.fix(length)
        m.fs.unit.dP_dx.fix(-membrane_pressure_drop / length)

        # test statistics
        assert number_variables(m) == 141
        assert number_total_constraints(m) == 111
        assert number_unused_variables(m) == 0

        # test degrees of freedom
        assert degrees_of_freedom(m) == 0

        # test scaling
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1, index=('Liq', 'H2O'))
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'NaCl'))

        calculate_scaling_factors(m)

        # check that all variables have scaling factors.
        # TODO: see aforementioned TODO on revisiting scaling and associated testing for property models.
        unscaled_var_list = list(unscaled_variables_generator(m.fs.unit, include_fixed=True))
        assert len(unscaled_var_list) == 0

        # check that all constraints have been scaled
        unscaled_constraint_list = list(unscaled_constraints_generator(m))
        assert len(unscaled_constraint_list) == 0

        # test initialization
        initialization_tester(m, fail_on_warning=True)

        # test variable scaling
        badly_scaled_var_lst = list(badly_scaled_var_generator(m))
        assert badly_scaled_var_lst == []

        # test solve
        results = solver.solve(m, tee=True)

        # Check for optimal solution
        assert_optimal_termination(results)

        # test solution
        assert (pytest.approx(-3.000e5, rel=1e-3) == value(m.fs.unit.deltaP[0]))
        assert (pytest.approx(4.562e-3, rel=1e-3) ==
                value(m.fs.unit.flux_mass_phase_comp_avg[0, 'Liq', 'H2O']))
        assert (pytest.approx(1.593e-6, rel=1e-3) ==
                value(m.fs.unit.flux_mass_phase_comp_avg[0, 'Liq', 'NaCl']))
        assert (pytest.approx(0.2281, rel=1e-3) ==
                value(m.fs.unit.permeate_side.properties_mixed[0].flow_mass_phase_comp['Liq', 'H2O']))
        assert (pytest.approx(7.963e-5, rel=1e-3) ==
                value(m.fs.unit.permeate_side.properties_mixed[0].flow_mass_phase_comp['Liq', 'NaCl']))
        assert (pytest.approx(41.96, rel=1e-3) ==
                value(m.fs.unit.feed_side.properties_interface_in[0].conc_mass_phase_comp['Liq', 'NaCl']))
        assert (pytest.approx(46.57, rel=1e-3) ==
                value(m.fs.unit.feed_side.properties_out[0].conc_mass_phase_comp['Liq', 'NaCl']))
        assert (pytest.approx(49.94, rel=1e-3) ==
                value(m.fs.unit.feed_side.properties_interface_out[0].conc_mass_phase_comp['Liq', 'NaCl']))

    @pytest.mark.unit
    def test_report(self, RO_frame):
        RO_frame.fs.unit.report()



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""1D reverse osmosis examples"""

from pyomo.environ import ConcreteModel
from idaes.core import FlowsheetBlock
from idaes.core.util.scaling import calculate_scaling_factors
from watertap.examples.flowsheets.full_treatment_train.model_components import property_models
from watertap.unit_models.reverse_osmosis_1D import (ReverseOsmosis1D,
                                                       ConcentrationPolarizationType,
                                                       MassTransferCoefficient,
                                                       PressureChangeType)
from watertap.examples.flowsheets.full_treatment_train.util import solve_with_user_scaling, check_dof


def build_RO(m, base='TDS', level='simple', name_str='RO'):
    """
    Builds a 1DRO model at a specified level (simple or detailed).
    Requires prop_TDS property package.
    """
    if base not in ['TDS']:
        raise ValueError('Unexpected property base {base} for build_RO'
                         ''.format(base=base))
    prop = property_models.get_prop(m, base=base)

    if level == 'simple':
        raise ValueError('Unexpected RO level {level} for build_RO'
                         ''.format(level=level))

    elif level == 'detailed':
        # build unit
        setattr(m.fs, name_str, ReverseOsmosis1D(default={
            "property_package": prop,
            "has_pressure_change": True,
            "pressure_change_type": PressureChangeType.calculated,
            "mass_transfer_coefficient": MassTransferCoefficient.calculated,
            "concentration_polarization_type": ConcentrationPolarizationType.calculated,
            "transformation_scheme": "BACKWARD",
            "transformation_method": "dae.finite_difference",
            "finite_elements": 10,
        }))
        blk = getattr(m.fs, name_str)

        # specify unit
        blk.area.fix(50)
        blk.A_comp.fix(4.2e-12)
        blk.B_comp.fix(3.5e-8)
        blk.mixed_permeate[0].pressure.fix(101325)
        blk.channel_height.fix(1e-3)
        blk.spacer_porosity.fix(0.97)
        blk.N_Re[0, 0].fix(500)

    else:
        raise ValueError('Unexpected argument {level} for level in build_RO'
                         ''.format(level=level))


def solve_RO(base='TDS', level='simple'):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    property_models.build_prop(m, base='TDS')

    build_RO(m, base=base, level=level)

    # specify feed
    property_models.specify_feed(m.fs.RO.feed_side.properties[0, 0], base='TDS')
    m.fs.RO.feed_side.properties[0, 0].pressure.fix(50e5)

    # scaling
    calculate_scaling_factors(m)

    # initialize
    m.fs.RO.initialize(optarg={'nlp_scaling_method': 'user-scaling'})

    m.fs.RO.display()
    check_dof(m)
    solve_with_user_scaling(m)

    m.fs.RO.report()

    return m


if __name__ == "__main__":
    # solve_RO(base='TDS', level='simple')
    solve_RO(base='TDS', level='detailed')



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

import pytest
import watertap.property_models.NaCl_prop_pack as props
from idaes.generic_models.properties.tests.test_harness import \
    PropertyTestHarness as PropertyTestHarness_idaes
from watertap.property_models.tests.property_test_harness import \
    (PropertyTestHarness, PropertyRegressionTest, PropertyCalculateStateTest)

# -----------------------------------------------------------------------------
@pytest.mark.unit
class TestNaClProperty_idaes(PropertyTestHarness_idaes):
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}
        self.prop_args = {}
        self.has_density_terms = False


class TestNaClProperty(PropertyTestHarness):
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}
        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e2}
        self.stateblock_statistics = {'number_variables': 20,
                                      'number_total_constraints': 16,
                                      'number_unused_variables': 1,  # pressure is unused
                                      'default_degrees_of_freedom': 3}  # 4 state vars, but pressure is not active
        self.default_solution = {('mass_frac_phase_comp', ('Liq', 'H2O')): 0.965,
                                 ('mass_frac_phase_comp', ('Liq', 'NaCl')): 0.035,
                                 ('dens_mass_phase', 'Liq'): 1021.5,
                                 ('flow_vol_phase', 'Liq'): 9.790e-4,
                                 ('conc_mass_phase_comp', ('Liq', 'H2O')): 985.7,
                                 ('conc_mass_phase_comp', ('Liq', 'NaCl')): 35.75,
                                 ('flow_mol_phase_comp', ('Liq', 'H2O')): 53.57,
                                 ('flow_mol_phase_comp', ('Liq', 'NaCl')): 0.5989,
                                 ('mole_frac_phase_comp', ('Liq', 'H2O')): 0.9889,
                                 ('mole_frac_phase_comp', ('Liq', 'NaCl')): 1.106e-2,
                                 ('molality_comp', 'NaCl'): 0.6206,
                                 ('diffus_phase', 'Liq'): 1.472e-9,
                                 ('visc_d_phase', 'Liq'): 1.055e-3,
                                 ('osm_coeff', None): 0.9271,
                                 ('pressure_osm', None): 2.853e6,
                                 ('enth_mass_phase', 'Liq'): 9.974e4}

@pytest.mark.component
class TestNaClPropertySolution_1(PropertyRegressionTest):
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.solver = 'ipopt'
        self.optarg = {'nlp_scaling_method': 'user-scaling'}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e2}
        self.state_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 0.95,
                           ('flow_mass_phase_comp', ('Liq', 'NaCl')): 0.05,
                           ('temperature', None): 273.15 + 25,
                           ('pressure', None): 50e5}
        self.regression_solution = {('mass_frac_phase_comp', ('Liq', 'H2O')): 0.95,
                                    ('mass_frac_phase_comp', ('Liq', 'NaCl')): 0.05,
                                    ('dens_mass_phase', 'Liq'): 1032.8,
                                    ('flow_vol_phase', 'Liq'): 9.682e-4,
                                    ('conc_mass_phase_comp', ('Liq', 'H2O')): 981.1,
                                    ('conc_mass_phase_comp', ('Liq', 'NaCl')): 51.64,
                                    ('flow_mol_phase_comp', ('Liq', 'H2O')): 52.73,
                                    ('flow_mol_phase_comp', ('Liq', 'NaCl')): 0.8556,
                                    ('mole_frac_phase_comp', ('Liq', 'H2O')): 0.9840,
                                    ('mole_frac_phase_comp', ('Liq', 'NaCl')): 1.597e-2,
                                    ('molality_comp', 'NaCl'): 0.9006,
                                    ('diffus_phase', 'Liq'): 1.471e-9,
                                    ('visc_d_phase', 'Liq'): 1.0875e-3,
                                    ('osm_coeff', None): 0.9347,
                                    ('pressure_osm', None): 4.174e6,
                                    ('enth_mass_phase', 'Liq'): 9.752e4}

@pytest.mark.component
class TestNaClPropertySolution_2(PropertyRegressionTest):
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.solver = 'ipopt'
        self.optarg = {'nlp_scaling_method': 'user-scaling'}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e2}
        self.state_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 0.74,
                           ('flow_mass_phase_comp', ('Liq', 'NaCl')): 0.26,
                           ('temperature', None): 273.15 + 25,
                           ('pressure', None): 50e5}
        self.regression_solution = {('mass_frac_phase_comp', ('Liq', 'H2O')): 0.74,
                                    ('mass_frac_phase_comp', ('Liq', 'NaCl')): 0.26,
                                    ('dens_mass_phase', 'Liq'): 1192,
                                    ('flow_vol_phase', 'Liq'): 8.392e-4,
                                    ('conc_mass_phase_comp', ('Liq', 'H2O')): 881.8,
                                    ('conc_mass_phase_comp', ('Liq', 'NaCl')): 309.8,
                                    ('flow_mol_phase_comp', ('Liq', 'H2O')): 41.08,
                                    ('flow_mol_phase_comp', ('Liq', 'NaCl')): 4.449,
                                    ('mole_frac_phase_comp', ('Liq', 'H2O')): 0.9022,
                                    ('mole_frac_phase_comp', ('Liq', 'NaCl')): 9.773e-2,
                                    ('molality_comp', 'NaCl'): 6.012,
                                    ('diffus_phase', 'Liq'): 1.580e-9,
                                    ('visc_d_phase', 'Liq'): 1.539e-3,
                                    ('osm_coeff', None): 1.274,
                                    ('pressure_osm', None): 3.796e7,
                                    ('enth_mass_phase', 'Liq'): 6.645e4}

@pytest.mark.component
class TestNaClPropertySolution_3(PropertyRegressionTest):
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.solver = 'ipopt'
        self.optarg = {'nlp_scaling_method': 'user-scaling'}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e2}
        self.state_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 0.999,
                           ('flow_mass_phase_comp', ('Liq', 'NaCl')): 0.001,
                           ('temperature', None): 273.15 + 25,
                           ('pressure', None): 10e5}
        self.regression_solution = {('mass_frac_phase_comp', ('Liq', 'H2O')): 0.999,
                                    ('mass_frac_phase_comp', ('Liq', 'NaCl')): 0.001,
                                    ('dens_mass_phase', 'Liq'): 995.8,
                                    ('flow_vol_phase', 'Liq'): 1.004e-3,
                                    ('conc_mass_phase_comp', ('Liq', 'H2O')): 994.8,
                                    ('conc_mass_phase_comp', ('Liq', 'NaCl')): 0.9958,
                                    ('flow_mol_phase_comp', ('Liq', 'H2O')): 55.45,
                                    ('flow_mol_phase_comp', ('Liq', 'NaCl')): 1.711e-2,
                                    ('mole_frac_phase_comp', ('Liq', 'H2O')): 0.9997,
                                    ('mole_frac_phase_comp', ('Liq', 'NaCl')): 3.084e-4,
                                    ('molality_comp', 'NaCl'): 1.713e-2,
                                    ('diffus_phase', 'Liq'): 1.508e-9,
                                    ('visc_d_phase', 'Liq'): 9.822e-4,
                                    ('osm_coeff', None): 0.918,
                                    ('pressure_osm', None): 7.797e4,
                                    ('enth_mass_phase', 'Liq'): 1.048e5}


@pytest.mark.component
class TestNaClCalculateState_1(PropertyCalculateStateTest):
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.solver = 'ipopt'
        self.optarg = {'nlp_scaling_method': 'user-scaling'}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1e-1,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e1}
        self.var_args = {('flow_vol_phase', 'Liq'): 2e-2,
                         ('mass_frac_phase_comp', ('Liq', 'NaCl')): 0.05,
                         ('temperature', None): 273.15 + 25,
                         ('pressure', None): 5e5}
        self.state_solution = {('flow_mass_phase_comp', ('Liq', 'H2O')): 19.62,
                               ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1.033}


@pytest.mark.component
class TestNaClCalculateState_2(PropertyCalculateStateTest):
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.solver = 'ipopt'
        self.optarg = {'nlp_scaling_method': 'user-scaling'}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1e1,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e3}
        self.var_args = {('flow_vol_phase', 'Liq'): 2e-4,
                         ('mole_frac_phase_comp', ('Liq', 'NaCl')): 0.05,
                         ('temperature', None): 273.15 + 25,
                         ('pressure', None): 5e5}
        self.state_solution = {('flow_mass_phase_comp', ('Liq', 'H2O')): 18.88e-2,
                               ('flow_mass_phase_comp', ('Liq', 'NaCl')): 3.224e-2}


@pytest.mark.component
class TestNaClCalculateState_3(PropertyCalculateStateTest):
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.solver = 'ipopt'
        self.optarg = {'nlp_scaling_method': 'user-scaling'}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e2}
        self.var_args = {('flow_vol_phase', 'Liq'): 1e-3,
                         ('pressure_osm', None): 100e5,
                         ('temperature', None): 273.15 + 25,
                         ('pressure', None): 5e5}
        self.state_solution = {('flow_mass_phase_comp', ('Liq', 'H2O')): 0.9608,
                               ('flow_mass_phase_comp', ('Liq', 'NaCl')): 0.1151}



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
'''
    This test is for the tutorials in the ../../tutorials/ directory.
    For now, it merely checks that the tutorials execute without raising
    an exception, it does not check for correctness
'''

import nbformat
import unittest
import glob
import os.path
from nbconvert.preprocessors import ExecutePreprocessor

_tutorials_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), '../../..', '..', 'tutorials')
_notebooks = glob.glob(os.path.join(_tutorials_dir, '**', '*.ipynb'), recursive=True)

class TestTutorials(unittest.TestCase):

    def test_tutorials(self):
        for notebook_filename in _notebooks:
            with self.subTest(notebook=os.path.basename(notebook_filename)):
                with open(notebook_filename) as f:
                    nb = nbformat.read(f, as_version=4)
                ep = ExecutePreprocessor()
                ep.preprocess(nb, 
                        {'metadata': {'path':f'{os.path.dirname(notebook_filename)}'}}) 

if __name__ == '__main__':
    unittest.main()



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
import pytest
from pyomo.environ import (ConcreteModel,
                           Block,
                           Var,
                           Constraint,
                           TerminationCondition,
                           SolverStatus,
                           value,
                           SolverFactory,
                           Expression,
                           TransformationFactory,
                           units as pyunits)
from pyomo.network import Arc, Port
from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.initialization import (solve_indexed_blocks,
                                            propagate_state)
from idaes.generic_models.unit_models import Mixer, Separator, Product, Feed
from idaes.generic_models.unit_models.mixer import MomentumMixingType
from pyomo.util.check_units import assert_units_consistent
from idaes.core.util.scaling import (unscaled_variables_generator,
                                     unscaled_constraints_generator)

import watertap.property_models.NaCl_prop_pack as props
from watertap.unit_models.reverse_osmosis_0D import ReverseOsmosis0D
from watertap.unit_models.pressure_exchanger import PressureExchanger
from watertap.unit_models.pump_isothermal import Pump
from watertap.examples.flowsheets.RO_with_energy_recovery.RO_with_energy_recovery import (
build, set_operating_conditions, initialize_system, solve, optimize_set_up, optimize, display_system, display_state, display_design)


solver = get_solver()

# -----------------------------------------------------------------------------
class TestEnergyRecoverySystem:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m = build()

        return m

    @pytest.mark.unit
    def test_build(self, system_frame):
        m = system_frame

        # model set up
        assert isinstance(m, ConcreteModel)
        assert isinstance(m.fs, FlowsheetBlock)
        assert isinstance(m.fs.properties, props.NaClParameterBlock)
        assert isinstance(m.fs.costing, Block)

        # unit models
        fs = m.fs
        assert isinstance(fs.feed, Feed)
        assert isinstance(fs.S1, Separator)
        assert isinstance(fs.P1, Pump)
        assert isinstance(fs.PXR, PressureExchanger)
        assert isinstance(fs.P2, Pump)
        assert isinstance(fs.M1, Mixer)
        assert isinstance(fs.RO, ReverseOsmosis0D)
        assert isinstance(fs.product, Product)
        assert isinstance(fs.disposal, Product)

        # unit model options
        # separator
        assert isinstance(fs.S1.P1, Port)
        assert isinstance(fs.S1.PXR, Port)
        # mixer
        assert isinstance(fs.M1.P1, Port)
        assert isinstance(fs.M1.P2, Port)
        assert isinstance(fs.M1.pressure_equality_constraints, Constraint)
        # RO
        assert isinstance(fs.RO.deltaP, Var)

        # additional expressions
        assert isinstance(fs.annual_water_production, Expression)
        assert isinstance(fs.specific_energy_consumption, Expression)

        # costing blocks
        blk_str_list = ['P1', 'P2', 'RO', 'PXR']
        for blk_str in blk_str_list:
            blk = getattr(fs, blk_str)
            c_blk = getattr(blk, 'costing')
            assert isinstance(c_blk, Block)
            assert isinstance(getattr(c_blk, 'capital_cost'), Var)
            assert isinstance(getattr(c_blk, 'operating_cost'), Var)

        var_str_list = ['capital_cost_total', 'investment_cost_total', 'operating_cost_MLC',
                        'operating_cost_total', 'LCOW']
        for var_str in var_str_list:
            var = getattr(fs.costing, var_str)
            assert isinstance(var, Var)

        # arcs
        arc_dict = {fs.s01: (fs.feed.outlet, fs.S1.inlet),
                    fs.s02: (fs.S1.P1, fs.P1.inlet),
                    fs.s03: (fs.P1.outlet, fs.M1.P1),
                    fs.s04: (fs.M1.outlet, fs.RO.inlet),
                    fs.s05: (fs.RO.permeate, fs.product.inlet),
                    fs.s06: (fs.RO.retentate, fs.PXR.high_pressure_inlet),
                    fs.s07: (fs.PXR.high_pressure_outlet, fs.disposal.inlet),
                    fs.s08: (fs.S1.PXR, fs.PXR.low_pressure_inlet),
                    fs.s09: (fs.PXR.low_pressure_outlet, fs.P2.inlet),
                    fs.s10: (fs.P2.outlet, fs.M1.P2)}
        for arc, port_tpl in arc_dict.items():
            assert arc.source is port_tpl[0]
            assert arc.destination is port_tpl[1]

        # units
        assert_units_consistent(fs)

    @pytest.mark.component
    def test_set_operating_conditions(self, system_frame):
        m = system_frame

        set_operating_conditions(m, water_recovery=0.5, over_pressure=0.3, solver=solver)

        # check fixed variables
        # feed
        assert m.fs.feed.pressure[0].is_fixed()
        assert value(m.fs.feed.pressure[0]) == 101325
        assert m.fs.feed.temperature[0].is_fixed()
        assert value(m.fs.feed.temperature[0]) == 298.15
        assert m.fs.feed.flow_mass_phase_comp[0, 'Liq', 'H2O'].is_fixed()
        assert value(m.fs.feed.flow_mass_phase_comp[0, 'Liq', 'H2O']) \
               == pytest.approx(0.9857, rel=1e-3)
        assert m.fs.feed.flow_mass_phase_comp[0, 'Liq', 'NaCl'].is_fixed()
        assert value(m.fs.feed.flow_mass_phase_comp[0, 'Liq', 'NaCl']) \
               == pytest.approx(3.575e-2, rel=1e-3)
        # pumps and pressure exchangers
        assert m.fs.P1.efficiency_pump[0].is_fixed()
        assert value(m.fs.P1.efficiency_pump[0]) == 0.8
        assert m.fs.P1.control_volume.properties_out[0].pressure.is_fixed()
        assert value(m.fs.P1.control_volume.properties_out[0].pressure) == pytest.approx(7.493e6, rel=1e-3)
        assert m.fs.P2.efficiency_pump[0].is_fixed()
        assert value(m.fs.P2.efficiency_pump[0]) == 0.8
        assert m.fs.PXR.efficiency_pressure_exchanger[0].is_fixed()
        assert value(m.fs.PXR.efficiency_pressure_exchanger[0]) == 0.95
        # RO
        assert m.fs.RO.A_comp[0, 'H2O'].is_fixed()
        assert value(m.fs.RO.A_comp[0, 'H2O']) == 4.2e-12
        assert m.fs.RO.B_comp[0, 'NaCl'].is_fixed()
        assert value(m.fs.RO.B_comp[0, 'NaCl']) == 3.5e-8
        assert m.fs.RO.channel_height.is_fixed()
        assert value(m.fs.RO.channel_height) == 1e-3
        assert m.fs.RO.spacer_porosity.is_fixed()
        assert value(m.fs.RO.spacer_porosity) == 0.97
        assert m.fs.RO.permeate.pressure[0].is_fixed()
        assert value(m.fs.RO.permeate.pressure[0]) == 101325
        assert m.fs.RO.width.is_fixed()
        assert value(m.fs.RO.width) == 5
        assert not m.fs.RO.area.is_fixed()
        assert value(m.fs.RO.area) == pytest.approx(50, rel=1e-3)

        # check degrees of freedom
        assert degrees_of_freedom(m) == 0

    @pytest.mark.component
    def test_initialize_system(self, system_frame):
        m = system_frame

        initialize_system(m, solver=solver)

        # check results across pressure exchanger, proxy for both upstream and downstream of RO
        # high pressure inlet
        assert value(m.fs.PXR.high_pressure_inlet.flow_mass_phase_comp[0, 'Liq', 'H2O']) \
               == pytest.approx(0.4928, rel=1e-3)
        assert value(m.fs.PXR.high_pressure_inlet.flow_mass_phase_comp[0, 'Liq', 'NaCl']) \
               == pytest.approx(3.561e-2, rel=1e-3)
        assert value(m.fs.PXR.high_pressure_inlet.temperature[0]) == pytest.approx(298.15, rel=1e-3)
        assert value(m.fs.PXR.high_pressure_inlet.pressure[0]) == pytest.approx(7.394e6, rel=1e-3)
        # low pressure inlet
        assert value(m.fs.PXR.low_pressure_inlet.flow_mass_phase_comp[0, 'Liq', 'H2O']) \
               == pytest.approx(0.4980, rel=1e-3)
        assert value(m.fs.PXR.low_pressure_inlet.flow_mass_phase_comp[0, 'Liq', 'NaCl']) \
               == pytest.approx(1.806e-2, rel=1e-3)
        assert value(m.fs.PXR.low_pressure_inlet.temperature[0]) == pytest.approx(298.15, rel=1e-3)
        assert value(m.fs.PXR.low_pressure_inlet.pressure[0]) == pytest.approx(101325, rel=1e-3)
        # low pressure outlet
        assert value(m.fs.PXR.low_pressure_outlet.pressure[0]) == pytest.approx(7.030e6, rel=1e-3)

    @pytest.mark.component
    def test_simulation(self, system_frame):
        m = system_frame

        solve(m, solver=solver)

        # check system metrics
        assert value(m.fs.RO.recovery_vol_phase[0, 'Liq']) == pytest.approx(0.4954, rel=1e-3)
        assert value(m.fs.specific_energy_consumption) == pytest.approx(2.727, rel=1e-3)
        assert value(m.fs.costing.LCOW) == pytest.approx(0.4394, rel=1e-3)

        # check mass balance
        assert (pytest.approx(value(m.fs.feed.outlet.flow_mass_phase_comp[0, 'Liq', 'H2O']), rel=1e-3)
                == value(m.fs.product.inlet.flow_mass_phase_comp[0, 'Liq', 'H2O'])
                + value(m.fs.disposal.inlet.flow_mass_phase_comp[0, 'Liq', 'H2O']))
        assert (pytest.approx(value(m.fs.feed.outlet.flow_mass_phase_comp[0, 'Liq', 'NaCl']), rel=1e-3)
                == value(m.fs.product.inlet.flow_mass_phase_comp[0, 'Liq', 'NaCl'])
                + value(m.fs.disposal.inlet.flow_mass_phase_comp[0, 'Liq', 'NaCl']))

    @pytest.mark.component
    def test_display_system(self, system_frame, capsys):
        m = system_frame
        display_system(m)

        captured = capsys.readouterr()

        assert captured.out == \
"""---system metrics---
Feed: 1.02 kg/s, 35000 ppm
Product: 0.493 kg/s, 280 ppm
Volumetric recovery: 49.5%
Water recovery: 50.0%
Energy Consumption: 2.7 kWh/m3
Levelized cost of water: 0.44 $/m3
"""

    @pytest.mark.component
    def test_display_design(self, system_frame, capsys):
        m = system_frame
        display_design(m)

        captured = capsys.readouterr()

        assert captured.out == \
"""---decision variables---
Operating pressure 74.9 bar
Membrane area 60.2 m2
---design variables---
Separator
Split fraction 50.53
Pump 1
outlet pressure: 74.9 bar
power 4.57 kW
Pump 2
outlet pressure: 74.9 bar
power 0.30 kW
"""


    @pytest.mark.component
    def test_display_state(self, system_frame, capsys):
        m = system_frame
        display_state(m)

        captured = capsys.readouterr()

        assert captured.out == \
"""---state---
Feed      : 1.021 kg/s, 35000 ppm, 1.0 bar
Split 1   : 0.505 kg/s, 35000 ppm, 1.0 bar
P1 out    : 0.505 kg/s, 35000 ppm, 74.9 bar
Split 2   : 0.516 kg/s, 35000 ppm, 1.0 bar
PXR LP out: 0.516 kg/s, 35000 ppm, 70.3 bar
P2 out    : 0.516 kg/s, 35000 ppm, 74.9 bar
Mix out   : 1.021 kg/s, 35000 ppm, 74.9 bar
RO perm   : 0.493 kg/s, 280 ppm, 1.0 bar
RO reten  : 0.528 kg/s, 67389 ppm, 73.9 bar
PXR HP out: 0.528 kg/s, 67389 ppm, 1.0 bar
"""

    @pytest.mark.component
    def test_optimization(self, system_frame):
        m = system_frame

        optimize_set_up(m)
        optimize(m, solver=solver)

        # check decision variables
        assert value(m.fs.RO.inlet.pressure[0]) == pytest.approx(5.708e6, rel=1e-3)
        assert value(m.fs.RO.area) == pytest.approx(115, rel=1e-3)
        # check system metrics
        assert value(m.fs.RO.recovery_vol_phase[0, 'Liq']) == pytest.approx(0.4954, rel=1e-3)
        assert value(m.fs.specific_energy_consumption) == pytest.approx(2.110, rel=1e-3)
        assert value(m.fs.costing.LCOW) == pytest.approx(0.4111, rel=1e-3)



"""
Error classes and utilities for the electrolyte database (EDB).
"""
__author__ = "Dan Gunter"

from typing import Dict
from pprint import pformat


class Error(Exception):
    """Abstract base class for all EDB errors."""
    pass


class ConfigGeneratorError(Error):
    """Base class of errors for ConfigGenerator actions and effects."""
    pass


class DataWrapperError(Error):
    """"Base class of errors for DataWrapper actions and effects."""
    pass


class BadConfiguration(DataWrapperError):
    """Bad configuration provided to build a DataWrapper class.
    """
    def __init__(self, whoami: str, config: Dict, missing: str = None, why: str = None):
        if missing:
            why = f"Missing field '{missing}'"
        elif why is None:
            why = "Unknown error"
        dumped = pformat(config)
        msg = f"Bad configuration provided to '{whoami}': {why}.\n{dumped}"
        super().__init__(msg)


class ValidationError(Error):
    """Validation error.
    """
    def __init__(self, err):
        msg = f"{err}"
        super().__init__(msg)


